import { createFlow, listenerChain, createEffect, getKey, delay, FLOW_UPDATE_EVENT, } from "./main";
export function retry(times, fn, ...args) {
    return (function* () {
        for (let i = 0; i < times; i++) {
            try {
                const result = yield fn(...args);
                return result;
            }
            catch (e) { }
        }
    })();
}
export function remove(key) {
    return createEffect((ec) => {
        if (Array.isArray(key)) {
            key.forEach((k) => ec.controller.remove(k));
        }
        else {
            ec.controller.remove(key);
        }
    });
}
export function debounce(ms) {
    return createEffect((ec) => {
        ec.call(function* () {
            // cancel previous
            yield cancel();
            // delay in specified time
            yield delay(ms);
        });
    });
}
export function cancel(target) {
    const hasTarget = !!arguments.length;
    return createEffect((ec) => {
        var _a;
        if (hasTarget) {
            // cancel flow by key
            if (typeof target === "function" || typeof target === "string") {
                ec.controller.cancelFlow(target);
            }
            // cancel cancellable object
            else if (typeof target === "object" &&
                typeof target.cancel === "function") {
                target === null || target === void 0 ? void 0 : target.cancel();
            }
            else {
                // throw error ?
            }
        }
        else {
            // cancel previous
            (_a = ec.flow.previous) === null || _a === void 0 ? void 0 : _a.cancel();
        }
        ec.next();
    });
}
export function stale(type, key, check) {
    return createEffect((ec) => {
        var _a;
        // stale('when', timeout)
        if (type === "when" && !isNaN(key)) {
            ec.flow.on("end", () => {
                let called = ec.flow.called;
                setTimeout(() => {
                    if (called !== ec.flow.called)
                        return;
                    ec.flow.stale = true;
                }, key);
            });
        }
        // stale('when', promise)
        else if (type === "when" && key && typeof key.then === "function") {
            ec.flow.on("end", () => {
                const called = ec.flow.called;
                const onDone = () => {
                    if (called !== ec.flow.called)
                        return;
                    ec.flow.stale = true;
                };
                if (typeof key.finally === "function") {
                    key.finally(onDone);
                }
                else {
                    key.then(onDone, onDone);
                }
            });
        }
        else {
            const events = type === "flow"
                ? [FLOW_UPDATE_EVENT]
                : Array.isArray(key)
                    ? key
                    : [key];
            const keys = (_a = (type === "flow"
                ? Array.isArray(key)
                    ? key
                    : [key]
                : undefined)) === null || _a === void 0 ? void 0 : _a.map(getKey);
            ec.flow.on("end", () => {
                let called;
                listenerChain((cleanup) => (payload) => {
                    if (ec.flow.status === "running") {
                        return;
                    }
                    if (ec.flow.called !== called) {
                        cleanup();
                        return;
                    }
                    if (type === "flow") {
                        const flow = payload;
                        if (keys && !keys.includes(flow.key))
                            return;
                    }
                    if (check && !check(payload)) {
                        return;
                    }
                    cleanup();
                    ec.flow.stale = true;
                }, (listener, add) => {
                    called = ec.flow.called;
                    add(ec.controller.on(events, listener));
                });
            });
        }
        ec.next();
    });
}
export function partial(data, wait) {
    return createEffect((ec) => {
        ec.flow.partial(data, wait || false);
        if (wait) {
            ec.flow.setNext(ec.next);
        }
        else {
            ec.next();
        }
    });
}
export function on(event, ...args) {
    const hasFlow = args.length > 1;
    return createEffect((ec) => {
        let latest;
        const cancel = listenerChain(() => (payload) => {
            if (!hasFlow) {
                cancel === null || cancel === void 0 ? void 0 : cancel();
                return ec.next(payload);
            }
            latest = createFlow({
                controller: ec.controller,
                parent: ec.flow,
                previous: latest,
                key: {},
                fn: args[0],
                onError: ec.fail,
            }).start(payload, ...args.slice(1));
        }, (listener, add, cleanup) => {
            add(ec.flow.on("end", cleanup));
            add(ec.controller.on(event, listener));
        });
        if (hasFlow) {
            return ec.next({ cancel });
        }
    });
}
export function emit(event, payload) {
    return createEffect((ec) => {
        ec.controller.emit(event, payload);
        ec.next();
    });
}
export function all(targets) {
    return handleParallel("all", targets);
}
export function race(targets) {
    return handleParallel("race", targets);
}
export function call(fn, ...args) {
    return createEffect((ec) => {
        ec.call(fn, ...args);
    });
}
export function start(key, ...args) {
    return createEffect((ec) => {
        run(ec, "start", key, args);
    });
}
export function restart(key, ...args) {
    return createEffect((ec) => {
        run(ec, "start", key, args);
    });
}
export function fork(flow, ...args) {
    return createEffect((ec) => {
        const { cancel, start } = createFlow({
            controller: ec.controller,
            parent: ec.flow,
            key: {},
            fn: flow,
        });
        start(...args);
        return ec.next({ cancel });
    });
}
export function spawn(flow, ...args) {
    return createEffect((ec) => {
        const { cancel, start } = createFlow({
            controller: ec.controller,
            key: {},
            fn: flow,
        });
        start(...args);
        return ec.next({ cancel });
    });
}
export function extra(value) {
    const hasValue = arguments.length;
    return createEffect((ec) => {
        if (hasValue) {
            Object.assign(ec.flow.extra, value);
        }
        else {
            ec.next(ec.flow.extra);
        }
    });
}
export function data() {
    return createEffect((ec) => {
        var _a;
        ec.next((_a = ec.flow.previous) === null || _a === void 0 ? void 0 : _a.data);
    });
}
export function update(key, value) {
    return createEffect((ec) => {
        const flow = ec.controller.flow(key);
        if ((flow === null || flow === void 0 ? void 0 : flow.key) === ec.flow.key) {
            throw new Error("Cannot update running flow");
        }
        flow === null || flow === void 0 ? void 0 : flow.update(value);
        ec.next();
    });
}
function run(ec, method, key, args) {
    let flow;
    if (typeof args[0] === "function" && typeof key !== "function") {
        flow = ec.controller.flow(key, args[0]);
        // remove first args
        args = args.slice(1);
    }
    else {
        flow = ec.controller.flow(key);
    }
    // the effect is never done
    if (!flow)
        return;
    flow[method](...args);
    const onUpdate = () => {
        const current = flow.current;
        if (current.status === "running")
            return;
        if (current.status === "faulted") {
            return ec.fail(current.error);
        }
        ec.next(current.data);
    };
    if (flow.status !== "running" && flow.status !== "idle") {
        return onUpdate();
    }
    const cleanup = ec.controller.on(FLOW_UPDATE_EVENT, (x) => {
        if (x.key !== flow.key)
            return;
        cleanup();
        onUpdate();
    });
    ec.flow.on("end", cleanup);
}
function handleParallel(type, targets) {
    const entries = Object.entries(targets);
    const isArray = Array.isArray(targets);
    return createEffect((ec) => {
        let done = false;
        let count = 0;
        const flows = new Map();
        const results = isArray ? [] : {};
        const cancelOthers = (current) => {
            flows.forEach((flow, key) => {
                if (key === current)
                    return;
                flow.cancel();
            });
        };
        const onDone = (key, success, value) => {
            if (done)
                return;
            count++;
            if (!success) {
                done = true;
                cancelOthers(key);
                return ec.fail(value);
            }
            results[key] = value;
            if (type === "race") {
                done = true;
                cancelOthers(key);
                if (isArray) {
                    return value;
                }
                return ec.next(results);
            }
            else if (count >= entries.length) {
                done = true;
                ec.next(results);
            }
        };
        entries.forEach(([key, target]) => {
            const fn = function* () {
                yield target;
            };
            const flow = createFlow({
                controller: ec.controller,
                parent: ec.flow,
                key: {},
                fn,
                onSuccess: (data) => {
                    onDone(key, true, data);
                },
                onError: (error) => {
                    onDone(key, false, error);
                },
            });
            flows.set(key, flow);
            flow.start();
        });
    });
}
