export const FLOW_UPDATE_EVENT = "#flow";
export const GLOBAL_STATUS_EVENT = "#status";
/**
 * create flow controller
 */
export function createController({ context, initData = {}, initFlow, onData, onCompleted, onFaulted, } = {}) {
    let data = Object.assign({}, initData);
    let promise = Promise.resolve();
    let ready = true;
    let error;
    const flows = new Map();
    const emitter = createEmitter({ wildcard: "*", privateEventPrefix: "#" });
    const controller = {
        context,
        get ready() {
            return ready;
        },
        get error() {
            return error;
        },
        get promise() {
            return promise;
        },
        emit: emitter.emit,
        on: emitter.on,
        start(...args) {
            return run("start", args);
        },
        restart(...args) {
            return run("restart", args);
        },
        cancelFlow(key) {
            var _a;
            (_a = flows.get(getKey(key))) === null || _a === void 0 ? void 0 : _a.cancel();
        },
        replaceFlow(currentFlow, withFlow) {
            if (flows.get(currentFlow.key) !== currentFlow) {
                return;
            }
            flows.set(currentFlow.key, withFlow);
        },
        flow(key, fn) {
            const keyType = typeof key;
            if (keyType === "function") {
                fn = key;
            }
            key = getKey(key);
            // flow(key)
            if (keyType !== "function" && arguments.length === 1) {
                return flows.get(key);
            }
            let flow = flows.get(key);
            if (!flow) {
                let data = undefined;
                let status;
                let hasData = false;
                if (typeof key === "string" && key in initData) {
                    data = initData[key];
                    hasData = true;
                    status = "completed";
                }
                flow = createFlow({
                    controller,
                    key,
                    fn,
                    initData: data,
                    initStatus: status,
                    hasData,
                });
                flows.set(key, flow);
                if (hasData) {
                    emitter.emit(FLOW_UPDATE_EVENT, flow);
                }
            }
            return flow;
        },
        remove(key) {
            var _a;
            key = getKey(key);
            // contains wildcard
            if (typeof key === "string" && key.indexOf("*") !== -1) {
                // ends with
                if (key[0] === "*") {
                    key = key.substr(1);
                    flows.forEach((flow, k) => {
                        if (typeof k === "string" && k.endsWith(key)) {
                            flow.dispose();
                            flows.delete(k);
                        }
                    });
                }
                else {
                    // starts with
                    key = key.substr(0, key.length - 1);
                    flows.forEach((flow, k) => {
                        if (typeof k === "string" && k.startsWith(key)) {
                            flow.dispose();
                            flows.delete(k);
                        }
                    });
                }
            }
            else {
                (_a = flows.get(key)) === null || _a === void 0 ? void 0 : _a.dispose();
                flows.delete(key);
            }
        },
        flowUpdated(flow) {
            if (flows.get(flow.key) !== flow)
                return;
            if (flow.status === "completed") {
                if (onData && typeof flow.key === "string") {
                    if (data[flow.key] !== flow.data) {
                        data = Object.assign(Object.assign({}, data), { [flow.key]: flow.data });
                        onData(data);
                    }
                }
                onCompleted === null || onCompleted === void 0 ? void 0 : onCompleted(flow);
            }
            else if (flow.status === "faulted") {
                onFaulted === null || onFaulted === void 0 ? void 0 : onFaulted(flow);
            }
            emitter.emit(FLOW_UPDATE_EVENT, flow);
        },
    };
    function run(type, inputs) {
        const args = [];
        const key = getKey(inputs[0]);
        let fn;
        // preload(flow, ...args)
        if (typeof inputs[0] === "function") {
            fn = inputs[0];
            args.push(...inputs.slice(1));
        }
        else {
            // preload(key, flow, ...args)
            fn = inputs[1];
            if (Array.isArray(inputs[0])) {
                args.push(...inputs[0].slice(1));
            }
            else {
                args.push(...inputs.slice(2));
            }
        }
        return controller.flow(key, fn)[type](...args);
    }
    if (initFlow) {
        ready = false;
        promise = new Promise((resolve, reject) => {
            createFlow({
                controller,
                key: {},
                fn: initFlow,
                onSuccess: (data) => {
                    if (isPlainObject(data)) {
                        Object.assign(initData, data);
                    }
                    ready = true;
                    resolve();
                },
                onError: (e) => {
                    ready = true;
                    error = e;
                    reject(error);
                },
            }).start();
        });
    }
    return controller;
}
export function createFlow({ controller, parent, previous, key, fn, onSuccess, onError, initData, initStatus = "idle", hasData, extra = {}, }) {
    let stale = false;
    let status = initStatus;
    let error;
    let childError;
    let data = initData;
    let cancelled = false;
    let disposed = false;
    let called = 0;
    let currentNext;
    let cancelFn;
    const iteratorStack = [];
    const emitter = createEmitter();
    function cleanup() {
        previous = undefined;
        emitter.emit("end");
        emitter.dispose();
    }
    function isRunning() {
        return !disposed && !stale && !flow.cancelled && status === "running";
    }
    function statusChanged(newStatus, value, forceUpdate) {
        if (!forceUpdate && !isRunning())
            return;
        stale = false;
        status = newStatus;
        cancelFn = undefined;
        if (newStatus === "completed") {
            hasData = true;
            data = value;
            onSuccess === null || onSuccess === void 0 ? void 0 : onSuccess(data);
            cleanup();
        }
        else if (newStatus === "faulted") {
            error = value;
            onError === null || onError === void 0 ? void 0 : onError(error);
            parent === null || parent === void 0 ? void 0 : parent.onChildError(error);
            cleanup();
        }
        emitter.emit("update", flow);
        controller.flowUpdated(flow);
    }
    function iteratorDone(iterator, done, value) {
        if (!isRunning())
            return;
        if (done) {
            if (iteratorStack.length) {
                return iteratorNext(iteratorStack.pop(), value);
            }
            return statusChanged("completed", value, false);
        }
        return iteratorYield(iterator, value);
    }
    function iteratorThrow(iterator, error) {
        var _a;
        if (!isRunning())
            return;
        try {
            const { done, value } = ((_a = iterator.throw) === null || _a === void 0 ? void 0 : _a.call(iterator, error)) || {};
            // error handled
            return iteratorDone(iterator, done, value);
        }
        catch (e) {
            if (iteratorStack.length) {
                return iteratorThrow(iteratorStack.pop(), e);
            }
            // error is not handled
            return statusChanged("faulted", e, false);
        }
    }
    function interatorCall(iterator, next) {
        iteratorStack.push(iterator);
        return iteratorNext(next, undefined);
    }
    function iteratorYield(iterator, value) {
        if (!isRunning())
            return;
        cancelFn = undefined;
        if (value) {
            // yield iterator
            if (typeof value.next === "function") {
                return interatorCall(iterator, value);
            }
            // yield promise
            if (value && typeof value.then === "function") {
                cancelFn = value.cancel;
                return value.then((resolved) => {
                    iteratorNext(iterator, resolved);
                }, (rejected) => {
                    iteratorThrow(iterator, rejected);
                });
            }
            if (typeof value === "function") {
                try {
                    const c = called;
                    const ec = {
                        flow,
                        controller,
                        context: controller.context,
                        next: (value) => {
                            if (c !== called) {
                                return;
                            }
                            iteratorNext(iterator, value);
                        },
                        fail: (error) => {
                            if (c !== called) {
                                return;
                            }
                            iteratorThrow(iterator, error);
                        },
                        call: (next, ...args) => {
                            if (c !== called) {
                                return;
                            }
                            if (typeof next === "function") {
                                try {
                                    return iteratorYield(iterator, next(...args));
                                }
                                catch (e) {
                                    return iteratorThrow(iterator, e);
                                }
                            }
                            return interatorCall(iterator, next);
                        },
                        end: (result) => {
                            if (c !== called) {
                                return;
                            }
                            iteratorStack.length = 0;
                            iteratorDone(iterator, true, result);
                        },
                    };
                    return value(ec);
                }
                catch (e) {
                    return iteratorThrow(iterator, e);
                }
            }
        }
        throw new Error(`Unsupported yield ${typeof value}: ${value}`);
    }
    function iteratorNext(iterator, payload) {
        if (stale || flow.cancelled)
            return;
        if (childError) {
            return iteratorThrow(iterator, childError);
        }
        try {
            const { done, value } = iterator.next(payload);
            return iteratorDone(iterator, done, value);
        }
        catch (e) {
            iteratorThrow(iterator, e);
        }
    }
    const flow = {
        key,
        fn,
        controller,
        statusChanged,
        get extra() {
            return extra;
        },
        get called() {
            return called;
        },
        get completed() {
            return status === "completed";
        },
        get faulted() {
            return status === "faulted";
        },
        get running() {
            return status === "running";
        },
        get idle() {
            return status === "idle";
        },
        get hasData() {
            return hasData || false;
        },
        get current() {
            return controller.flow(key) || flow;
        },
        get parent() {
            return parent;
        },
        get previous() {
            return previous;
        },
        get data() {
            return data;
        },
        set data(value) {
            flow.update(value);
        },
        get status() {
            return status;
        },
        get error() {
            return error;
        },
        get stale() {
            return stale;
        },
        set stale(value) {
            // cannot turn stale to unstale
            if (stale)
                return;
            stale = value;
            controller.flowUpdated(flow);
            cleanup();
        },
        get cancelled() {
            return cancelled || (parent === null || parent === void 0 ? void 0 : parent.cancelled) || false;
        },
        on: emitter.on,
        cancel() {
            if (cancelled)
                return flow;
            cancelled = true;
            cancelFn === null || cancelFn === void 0 ? void 0 : cancelFn();
            emitter.emit("cancel", flow);
            cleanup();
            return flow;
        },
        partial(partialData, wait) {
            if (!isRunning())
                return flow;
            data = partialData;
            hasData = true;
            if (wait) {
                status = "completed";
            }
            controller.flowUpdated(flow);
            return flow;
        },
        start(...args) {
            // running or finished
            if (status !== "idle" && !stale) {
                return flow;
            }
            cancelFn = undefined;
            cancelled = false;
            stale = false;
            called++;
            if (flow.previous) {
                controller.replaceFlow(flow.previous, flow);
            }
            emitter.emit("start", flow);
            try {
                status = "running";
                const result = fn(...args);
                // promise
                if (result && typeof result.then === "function") {
                    cancelFn = result.cancel;
                    statusChanged("running", undefined, false);
                    result.then((resolved) => {
                        statusChanged("completed", resolved, false);
                    }, (rejected) => {
                        statusChanged("faulted", rejected, false);
                    });
                }
                // iterator
                else if (result && typeof result.next === "function") {
                    iteratorNext(result, undefined);
                    if (isRunning()) {
                        controller.flowUpdated(flow);
                    }
                }
                else {
                    statusChanged("completed", result, false);
                }
            }
            catch (e) {
                statusChanged("faulted", e, false);
            }
            return flow;
        },
        restart(...args) {
            if (status === "idle" || stale) {
                return flow.start(...args);
            }
            return createFlow({
                controller,
                parent,
                previous: flow,
                key,
                fn,
                onSuccess,
                onError,
                initData: data,
                hasData,
                extra,
            }).start(...args);
        },
        update(value) {
            if (status !== "running") {
                if (typeof value === "function") {
                    value = value(data);
                }
                statusChanged("completed", value, true);
            }
            return flow;
        },
        dispose() {
            if (disposed)
                return flow;
            disposed = true;
            cancelled = true;
            emitter.dispose();
            return flow;
        },
        onChildError(error) {
            childError = error;
        },
        setNext(value) {
            currentNext = value;
        },
        next(payload) {
            if (!cancelled && !stale && currentNext) {
                const next = currentNext;
                currentNext = undefined;
                status = "running";
                next === null || next === void 0 ? void 0 : next.call(null, payload);
                if (status === "running") {
                    controller.flowUpdated(flow);
                }
            }
            return flow;
        },
    };
    return flow;
}
export function makeKey(args) {
    return args.join(":");
}
export function getKey(value) {
    if (typeof value === "function") {
        return value.flowKey || value;
    }
    if (Array.isArray(value)) {
        return makeKey(value);
    }
    return value;
}
export function configure(fn, configs) {
    const { key } = (typeof configs === "string" ? { key: configs } : configs) || {};
    if (typeof key !== "undefined") {
        fn.flowKey = key;
    }
    return fn;
}
export function createEffect(fn) {
    return fn;
}
export function createEmitter({ wildcard, privateEventPrefix, } = {}) {
    const events = new Map();
    function getHandlers(event) {
        let handlers = events.get(event);
        if (!handlers) {
            handlers = [];
            events.set(event, handlers);
        }
        return handlers;
    }
    function on(...args) {
        if (args.length > 1) {
            const [event, listener] = args;
            const removes = (Array.isArray(event) ? event : [event]).map((event) => {
                const handlers = getHandlers(event);
                handlers.push(listener);
                let active = true;
                return () => {
                    if (!active)
                        return;
                    active = false;
                    const index = handlers.indexOf(listener);
                    if (index !== -1)
                        handlers.splice(index, 1);
                    return listener;
                };
            });
            if (Array.isArray(event)) {
                return () => removes.forEach((x) => x);
            }
            return removes[0];
        }
        let cancel = () => { };
        return Object.assign(new Promise((resolve) => {
            cancel = on(args[0], (payload) => {
                cancel();
                resolve(payload);
            });
        }), { cancel });
    }
    function emit(event, payload) {
        try {
            const handlers = getHandlers(event);
            handlers.slice(0).forEach((handler) => handler(payload));
        }
        finally {
            if (wildcard &&
                event !== wildcard &&
                (!privateEventPrefix ||
                    typeof event !== "string" ||
                    privateEventPrefix !== event[0])) {
                emit(wildcard, { type: event, payload });
            }
        }
    }
    function dispose() {
        events.forEach((listeners) => {
            listeners.length = 0;
        });
        events.clear();
    }
    return { emit, on, dispose };
}
export const delay = (ms = 0) => {
    let timeout;
    return Object.assign(new Promise((resolve) => (timeout = setTimeout(resolve, ms))), {
        cancel() {
            clearTimeout(timeout);
        },
    });
};
export function listenerChain(listenerFactory, register) {
    const removeListeners = [];
    const cleanup = () => {
        removeListeners.forEach((x) => x());
    };
    register(listenerFactory(cleanup), (x) => {
        removeListeners.push(x);
        return x;
    }, cleanup);
    return cleanup;
}
function isPlainObject(value) {
    return typeof value == "object" && value.constructor == Object;
}
